import { ErrorMapper } from 'utils/ErrorMapper';
import { runAsHarvester } from './creep/harvester';
import { runAsBuilder } from './creep/builder';
import { runAsUpgrader } from './creep/upgrader';
import { uniqId } from './utils/Identity';

function makeName(): string {
  return uniqId();
}

enum CreepRole {
  HARVESTER = 'harvester',
  UPGRADER = 'upgrader',
  BUILDER = 'builder'
}

function makeCreep(role: CreepRole, count: number) {
  const creeps = Object.values(Game.creeps).filter((c) => c.memory.role === role);
  console.log(`Make creep "${role}"`, 'need', count, 'has', creeps.length);
  if (creeps.length < count) {
    const firstSpawn = _.first(Object.values(Game.spawns));
    const name = makeName();
    const memory = { role: role, sourceId: '' } as CreepMemory;
    const err = firstSpawn.spawnCreep([WORK, WORK, CARRY, CARRY, MOVE, MOVE], name, {
      memory,
      directions: [BOTTOM, BOTTOM_RIGHT, RIGHT]
    });
    console.log(`Make creep "${role}"`, 'err', err);
  }
}

function callHarvestersFromOthers(minHarvCount: number) {
  const harvesters = Object.values(Game.creeps).filter((c) => c.memory.role === CreepRole.HARVESTER);
  if (harvesters.length < minHarvCount) {
    const others = Object.values(Game.creeps).filter((c) => c.memory.role !== CreepRole.HARVESTER);
    const required = Math.min(minHarvCount, others.length);
    console.log('Call harvesters', required);
    let count = 0;
    for (let creep of others) {
      if (count >= required) {
        break;
      }
      count++;
      creep.memory.role = CreepRole.HARVESTER;
    }
  }
}

// When compiling TS to JS and bundling with rollup, the line numbers and file names in error messages change
// This utility uses source maps to get the line numbers and file names of the original, TS source code
export const loop = ErrorMapper.wrapLoop(() => {
  console.log('');
  console.log(`Current game tick is ${Game.time}`);

  // Automatically delete memory of missing creeps
  for (const name in Memory.creeps) {
    if (!(name in Game.creeps)) {
      delete Memory.creeps[name];
    }
  }

  callHarvestersFromOthers(2);

  makeCreep(CreepRole.HARVESTER, 4);
  makeCreep(CreepRole.UPGRADER, 4);
  makeCreep(CreepRole.BUILDER, 4);

  // Process current creeps
  for (let name in Game.creeps) {
    const creep = Game.creeps[name];
    if (creep.memory.role === CreepRole.HARVESTER) {
      runAsHarvester(creep);
    }
    if (creep.memory.role === CreepRole.UPGRADER) {
      runAsUpgrader(creep);
    }
    if (creep.memory.role === CreepRole.BUILDER) {
      runAsBuilder(creep);
    }
  }
});
